import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# default mass is He molecule mass
def BM_pdf(v, T=300, m=6.648837e-27):
  k=1.38e-23
  return 4*np.pi*(m/(2*np.pi*k*T))**(3/2)*v**2*np.exp(-m*v**2/(2*k*T))

class particles():
  def __init__(self, N=100, T=300):
    # X, Y, Z, Vx, Vy, Vz
    self.N=N
    self.T=T
    self.bound=np.array(([0, 2], [0, 1], [0, 1]))
    self.time_step=0
    self.parti=np.zeros([N, 6])
    # random CDF build
    # set maximum velocity 6000 m/s
    v=np.linspace(0, 6000, 6000*100)
    pdf=BM_pdf(v, T)/100
    cdf=np.copy(pdf)
    for i in range(len(pdf)):
      cdf[i]+=cdf[i-1]
    
    # initial random state
    for i in range(N):
      self.parti[i, 0]=np.random.uniform(0, 1)
      self.parti[i, 1]=np.random.uniform(0, 1)
      self.parti[i, 2]=np.random.uniform(0, 1)
      
      self.parti[i, 3]=np.random.uniform(-0.1, 0.1)
      self.parti[i, 4]=np.random.uniform(-0.1, 0.1)
      self.parti[i, 5]=np.random.uniform(-0.1, 0.1)
      self.parti[i, 3:]/=np.linalg.norm(self.parti[i, 3:])
      uni=np.random.random()
      bm=np.where(cdf<uni)[0][-1]
      self.parti[i, 3:]*=v[bm]
  
  def velo_plot(self, time):
    velo_distri=[]
    for i in range(N):
      velo_distri.append(np.linalg.norm(self.parti[i, 3:]))
    plt.hist(velo_distri, bins=100)
    plt.pause(time)
    plt.close()
  
  def step(self):
    # bound scale 10km : 1
    scale=10000
    for i in range(self.N):
      for j in range(3):
        self.parti[i, j]+=self.parti[i, j+3]/scale
      for j in range(3):
        if self.parti[i, j]<=self.bound[j][0]:
          self.parti[i, j]=-self.parti[i, j]
          self.parti[i, j+3]=-self.parti[i, j+3]
        elif self.parti[i, j]>=self.bound[j][1]:
          self.parti[i, j]=self.bound[j][1]-(self.parti[i, j]-self.bound[j][1])
          self.parti[i, j+3]=-self.parti[i, j+3]
    self.time_step+=1
  
  def plot(self, state=0, time=0.1):
    fig=plt.figure(figsize=(9, 6))
    ax=fig.add_subplot(111, projection="3d")
    axes=[2, 1, 1]
    xyz=np.ones(axes, dtype=np.bool)
    colors=np.empty(axes+[4], dtype=np.float32); colors[:]=[1, 1, 0, 0.1]
    ax.voxels(xyz, facecolors=colors)
    ax.scatter(self.parti[:, 0], self.parti[:, 1], self.parti[:, 2], alpha=1, s=1/self.N*1000)
    ax.set_xlabel("x [0, 2]"); ax.set_ylabel("y [0, 1]"); ax.set_zlabel("z [0, 1]")
    ax.set_xticks([0, 2, 2]); ax.set_yticks([0, 1, 1]); ax.set_zticks([0, 1, 1])
    ax.set_xlabel('x [10 km]'); ax.set_ylabel('y [10 km]'); ax.set_zlabel('z [10 km]')
    
    plt.title("time step: %d" %self.time_step)
    if state:
      plt.show(block=False)
      plt.pause(time)
      plt.close()
    else:
      plt.show()

if __name__=="__main__":
  N=100
  T=300
  test=particles(N, T)
  # plot particles velocity distribution
  #test.velo_plot(time=5)
  
  num=50
  #num=input("Time step count: ")
  for i in range(int(num)):
    test.step()
    test.plot(state=1, time=0.1)
